@@ -79,6 +79,7 @@ class BaseAdapter(object):
     operations = {'eq': '= %s'}
     interpolation = '%s'
     sequence_support = False
+    for_update_support = False
     reserved_tables = []
     quote_char = '"'
     
@@ -185,6 +186,7 @@ class PostgresqlAdapter(BaseAdapter):
     }
     reserved_tables = ['user']
     sequence_support = True
+    for_update_support = True
 
     def connect(self, database, **kwargs):
         if not psycopg2:
@@ -231,6 +233,7 @@ class MySQLAdapter(BaseAdapter):
         'startswith': 'LIKE BINARY %s',
     }
     quote_char = '`'
+    for_update_support = True
 
     def connect(self, database, **kwargs):
         if not mysql:
@@ -1269,6 +1272,7 @@ class SelectQuery(BaseQuery):
         self._offset = None
         self._distinct = False
         self._qr = None
+        self._for_update = False
         super(SelectQuery, self).__init__(model)
     
     def clone(self):
@@ -1281,6 +1285,7 @@ class SelectQuery(BaseQuery):
         query._offset = self._offset
         query._distinct = self._distinct
         query._qr = self._qr
+        query._for_update = self._for_update
         query._where = self.clone_where()
         query._where_models = set(self._where_models)
         query._joined_models = self._joined_models.copy()
@@ -1302,8 +1307,12 @@ class SelectQuery(BaseQuery):
     def offset(self, num_rows):
         self._offset = num_rows
     
+    @returns_clone
+    def for_update(self, for_update=True):
+        self._for_update = for_update
+    
     def count(self):
-        if self._distinct:
+        if self._distinct or self._group_by:
             return self.wrapped_count()
         
         clone = self.order_by()
@@ -1536,6 +1545,9 @@ class SelectQuery(BaseQuery):
         if self._offset:
             pieces.append('OFFSET %d' % self._offset)
         
+        if self._for_update and self.database.adapter.for_update_support:
+            pieces.append('FOR UPDATE')
+        
         return ' '.join(pieces), params, query_meta
     
     def sql(self):
@@ -2230,6 +2242,12 @@ class BaseModelOptions(object):
                 if name is None or name == field.name or name == field.db_column:
                     return field
     
+    def get_field_for_related_name(self, model, related_name):
+        for field in model._meta.fields.values():
+            if isinstance(field, ForeignKeyField) and field.to == self.model_class:
+                if field.related_name == related_name:
+                    return field
+    
     def rel_exists(self, model):
         return self.get_related_field_for_model(model) or \
                self.get_reverse_related_field_for_model(model)
@@ -2435,7 +2453,60 @@ class Model(object):
             new_pk = insert.execute()
             setattr(self, self._meta.pk_name, new_pk)
 
-    def delete_instance(self):
+    @classmethod
+    def collect_models(cls, accum=None):
+        # dfs to grab any affected models, then from the bottom up issue
+        # proper deletes using subqueries to obtain objects to remove
+        accum = accum or []
+        models = []
+        
+        for related_name, rel_model in cls._meta.reverse_relations.items():
+            rel_field = cls._meta.get_field_for_related_name(rel_model, related_name)
+            coll = [(rel_model, rel_field.name, rel_field.null)] + accum
+            if not rel_field.null:
+                models.extend(rel_model.collect_models(coll))
+            
+            models.append(coll)
+        return models
+    
+    def collect_queries(self):
+        select_queries = []
+        nullable_queries = []
+        collected_models = self.collect_models()
+        if collected_models:
+            for model_joins in collected_models:
+                depth = len(model_joins)
+                base, last, nullable = model_joins[0]
+                query = base.select([base._meta.pk_name])
+                for model, join, _ in model_joins[1:]:
+                    query = query.join(model, on=last)
+                    last = join
+                
+                query = query.where(**{last: self.get_pk()})
+                if nullable:
+                    nullable_queries.append((query, last, depth))
+                else:
+                    select_queries.append((query, last, depth))
+        return select_queries, nullable_queries
+
+    def delete_instance(self, recursive=False):
+        # XXX: it is strongly recommended you run this in a transaction if using
+        # the recursive delete
+        if recursive:
+            # reverse relations, i.e. anything that would be orphaned, delete.
+            select_queries, nullable_queries = self.collect_queries()
+
+            for query, fk_field, depth in select_queries:
+                model = query.model
+                model.delete().where(**{
+                    '%s__in' % model._meta.pk_name: query,
+                }).execute()
+            for query, fk_field, depth in nullable_queries:
+                model = query.model
+                model.update(**{fk_field: None}).where(**{
+                    '%s__in' % model._meta.pk_name: query,
+                }).execute()
+        
         return self.delete().where(**{
             self._meta.pk_name: self.get_pk()
         }).execute()
